/* Copyright (c) 2000-2006 ActiveState Software Inc.
   See the file LICENSE.txt for licensing information. */

/* JavaScript-side support for Komodo's linting
 *
 * Usage:
 *  - This module is initialized at Komodo startup: lint_initialize().
 *  - A LintBuffer instance is lazily created for each view and attached
 *    to that view.
 *  - lintBuffer.request() is called to request a lint at some point
 *  - lintBuffer.reportResults() is called from the backend after processing
 *    a lint request
 *  - lintBuffer.destructor() is called when destroying it
 *  - This module is finalized at Komodo shutdown: lint_finalize().
 *
 * Notes on the various states of lint requests:
 * - Requests are made to lint when a view changes in some significant way
 *   (content changed, language changed, etc.). For optimization reasons
 *   a lint request is not made of the backend until a certain delay has
 *   expired. Until then the lint requests is known as a "delayed" request.
 *   The back end is basically a queue of lint requests and a thread
 *   processing them one at a time. If a lint request is in the queue it
 *   is known as a "pending" request. If the thread is currently processing
 *   a request it is known as the "current" request.
 */

var gLintDisplayer = null;
var gLintObserverSvc = null;
var gLintService = null;
var gLintLog = null;
// gPrefs is expected to be defined externally.



//---- The new LintBuffer class (replacement for jsLintBuffer)
//
// There is one of these for each Komodo 'editor' view. It is responsible
// for all JS-side linting handling for that view.
//

function LintBuffer(view) {
    gLintLog.info("LintBuffer["+view.title+"].constructor()");
    try {
        this.view = view;
        this.lintingEnabled = this.view.prefs.getBooleanPref("editUseLinting");
        this.lintResults = null;
        this.errorString = null;
        this._lastRequestId = 0; // used to ensure only the last request is used

        gPrefs.prefObserverService.addObserver(this, "editUseLinting", false);
        gPrefs.prefObserverService.addObserver(this, "perl_lintOption", false);
        this.view.prefs.prefObserverService.addObserver(this,
                                                        "editUseLinting",
                                                        false);

        this._lintTimer = null; // used to control when lint requests are issued
    } catch(ex) {
        gLintLog.exception(ex);
    }
}
LintBuffer.prototype.constructor = LintBuffer;


LintBuffer.prototype.QueryInterface = function (iid)
{
    if (!iid.equals(Components.interfaces.nsIObserver) &&
        !iid.equals(Components.interfaces.koILintBuffer) &&
        !iid.equals(Components.interfaces.nsISupports)) {
        throw Components.results.NS_ERROR_NO_INTERFACE;
    }
    return this;
}


LintBuffer.prototype.destructor = function()
{
    gLintLog.info("LintBuffer["+this.view.title+"].destructor()");
    try {
        gLintService.cancelPendingRequests(this.view.uid);
        this._cancelDelayedRequest();
        this._clearResults();

        this.view.prefs.prefObserverService.removeObserver(this,
                                                           "editUseLinting");
        gPrefs.prefObserverService.removeObserver(this, "editUseLinting");
        gPrefs.prefObserverService.removeObserver(this, "perl_lintOption");

        this.view = null; // drop reference to the view
    } catch(ex) {
        gLintLog.exception(ex);
    }
}


// nsIObserver.observe()
LintBuffer.prototype.observe = function(subject, topic, data)
{
    //gLintLog.debug("LintBuffer["+this.view.title+"].observe: subject="+
    //               subject+", topic="+topic+", data="+data);
    try {
        switch (topic) {
        case "perl_lintOption":
            lintingEnabled = this.view.prefs.getBooleanPref("editUseLinting");
            gLintLog.info("LintBuffer["+this.view.title+
                           "].observe: lintingEnabled="+lintingEnabled);
            if (lintingEnabled && this.view.languageObj.name == "Perl") {
                this.request();
            }
            break;
        case "editUseLinting":
            lintingEnabled = this.view.prefs.getBooleanPref("editUseLinting");
            gLintLog.info("LintBuffer["+this.view.title+
                           "].observe: lintingEnabled="+lintingEnabled);
            if (lintingEnabled != this.lintingEnabled) {
                // Do whatever must be done when the lintingEnabled state changes.
                this.lintingEnabled = lintingEnabled;
                if (lintingEnabled) {
                    this.request();
                } else {
                    gLintService.cancelPendingRequests(this.view.uid);
                    this._cancelDelayedRequest();
                    this._clearResults();
                    this._notify();
                }
            }
        }
    } catch(ex) {
        gLintLog.exception(ex);
    }
}


// Called to request that the view be linted sometime soon.
//
// Note: To attempt to avoid a flurry of lints when the user is typing into
// the document (because lints can be expensive) every .request() does not
// necessarily result in a lint. Lint requests are only forwarded to the
// linting backend after a delay 
//
LintBuffer.prototype.request = function(reason /* = "" */)
{
    if (typeof(reason) == "undefined" || reason == null) reason = "";
    gLintLog.info("LintBuffer["+this.view.title+"].request(reason='"+
                  reason+"')");
    try {
        gLintService.cancelPendingRequests(this.view.uid);
        this._clearResults();

        // avoid linting zero length documents
        if (typeof(this.view.scimoz) != 'undefined' &&
            this.view.scimoz.length < 1) {
            return;
        }

        // Increment this here instead of in ._issueRequest() to ensure that
        // a request issued at time T1 is not considered current when its
        // results are reported at time T2 if the buffer has changed between
        // those two times.
        this._lastRequestId += 1;

        // cancel the pending timeout if the request comes
        // in before the request timeout has expired
        if (this._lintTimer) {
            this._lintTimer.stopTimeout();
        }
        this._lintTimer = new objectTimer(this, this._issueRequest, []);
        var delay;
        if (! gPrefs.getBooleanPref('editUseLinting')) {
            delay = 0;
        } else {
            delay = gPrefs.getLongPref('lintDelay'); // lint request delay (in ms)
        }
        this._lintTimer.startTimeout(delay);

        this._notify();
    } catch(ex) {
        gLintLog.exception(ex);
    }
}


// koILintBuffer.reportResults()
LintBuffer.prototype.reportResults = function(request)
{
    gLintLog.info("LintBuffer["+this.view.title+"].reportResults(request)");
    try {
        // Ignore results that are not the last one issued (it means there
        // is another one in the pipe).
        if (request.rid == this._lastRequestId) {
            this.lintResults = request.results;
            this.errorString = request.errorString;
            if (this.lintResults) {
                gLintDisplayer.display(this.view.scimoz, this.lintResults,
                                       this.view.document.languageObj.indicatorBits,
                                       0, this.view.scimoz.length);
            }
            this._notify();
        }
    } catch(ex) {
        gLintLog.exception(ex);
    }
}


// Actually issue a lint request to the linting backend.
LintBuffer.prototype._issueRequest = function()
{
    gLintLog.info("LintBuffer["+this.view.title+"]._issueRequest()");
    try {
        var lr = this._createLintRequest(this._getLinterCID());
        gLintService.addRequest(lr);
        this._cancelDelayedRequest();
    } catch(ex) {
        gLintLog.exception(ex);
    }
}


// Cancel the delayed request if there is one.
LintBuffer.prototype._cancelDelayedRequest = function()
{
    gLintLog.debug("LintBuffer["+this.view.title+"]._cancelDelayedRequest()");
    try {
        if (this._lintTimer) {
            this._lintTimer.stopTimeout();
            this._lintTimer.free();
            this._lintTimer = null;
        }
    } catch(ex) {
        gLintLog.exception(ex);
    }
}


// Clear the current lint results, if there are any.
LintBuffer.prototype._clearResults = function()
{
    gLintLog.debug("LintBuffer["+this.view.title+"]._clearResults()");
    try {
        if (this.lintResults) {
            // .displayClear() has to be called in a timeout to avoid
            // re-entrance into scintilla: we are already mostly likely in
            // a scintilla onModified event handler (linting is most
            // common requested for a buffer modification).
            window.setTimeout(
                function(lintDisplayer, view) {
                    if (view && view.document) {
                        lintDisplayer.displayClear(view.scimoz,
                            view.document.languageObj.indicatorBits);
                    }
                },
                0, gLintDisplayer, this.view
            );
            this.lintResults = null;
        }
        this.errorString = null;
    } catch(ex) {
        gLintLog.exception(ex);
    }
}



// Notify of a change in this view's check status/state.
LintBuffer.prototype._notify = function()
{
    gLintLog.debug("LintBuffer["+this.view.title+"]._notify()");
    try {
        if (this.view == gViewMgr.currentView) {
            try {
                gLintObserverSvc.notifyObservers(
                    this.view, "current_view_check_status","");
            } catch(ex1) {
                // ignore
            }
        }
    } catch(ex2) {
        gLintLog.exception(ex2);
    }
}

// Create a new koILintRequest instance and return it.
LintBuffer.prototype._createLintRequest = function(linterType)
{
    gLintLog.debug("LintBuffer["+this.view.title+
                   "]._createLintRequest(linterType="+linterType+")");
    try {
        // Linters can't chdir to a remote URL, so don't pass in a working
        // directory unless it's local.
        var cwd = null;
        if (!this.view.document.isUntitled &&
            this.view.document.file.isLocal)
        {
            cwd = this.view.document.file.dirName;
        }

        var lr = Components.classes["@activestate.com/koLintRequest;1"].
                    createInstance(Components.interfaces.koILintRequest);
        lr.rid = this._lastRequestId;
        lr.document = this.view.document;
        lr.linterType = linterType;
        lr.uid = this.view.uid;
        lr.cwd = cwd;
        lr.lintBuffer = this;

        return lr;
    } catch(ex) {
        gLintLog.exception(ex);
    }
    return null;
}


// Return an appropriate koILinter contract ID for the current language.
// If there isn't one, return null.
LintBuffer.prototype._getLinterCID = function()
{
    gLintLog.debug("LintBuffer["+this.view.title+"]._getLinterCID()");
    try {
        var cid = null;
        var linterLanguageService = this.view.document.languageObj.
            getLanguageService(Components.interfaces.koILinterLanguageService);
        if (linterLanguageService) {
            cid = linterLanguageService.linterCID;
        }
        return cid;
    } catch(ex) {
        gLintLog.exception(ex);
    }
    return null;
}



//---- public lint interface

// Should be called in Komodo initialization.
function lint_initialize() {
    gLintLog = getLoggingMgr().getLogger("lint");
    //gLintLog.setLevel(LOG_INFO);
    gLintLog.info("lint_initialize()");
    try {
        gLintDisplayer = Components.classes["@activestate.com/koLintDisplayer;1"].
                         getService(Components.interfaces.koILintDisplayer);
        gLintService = Components.classes["@activestate.com/koLintService;1"].
                       getService(Components.interfaces.koILintService);
        gLintObserverSvc = Components.classes["@mozilla.org/observer-service;1"].
                           getService(Components.interfaces.nsIObserverService);

        KomodoRegisterOnUnloadHandler(lint_finalize);
    } catch(ex) {
        gLintLog.exception(ex);
    }
}


// This is registered to be called on Komodo shutdown by lint_initialize().
function lint_finalize() {
    gLintLog.info("lint_finalize()");
    try {
        if (gLintService) {
            // free the lint manager
            gLintService.terminate();
            gLintService = null;
        }
    } catch(ex) {
        gLintLog.exception(ex);
    }
}


function lint_jumpToNextLintResult()
{
    try {
        gLintLog.debug("lint_jumpToNextLintResult()");

        var view = gViewMgr.currentView;
        if (!view) return;
        if (typeof(view.lintBuffer) == "undefined") return;

        // If there are no results, then say so.
        //   lintResults == null        -> inprogress
        //   lintResults.length == 0    -> ok
        //   lintResults.length > 0     -> errors/warnings
        var lintResults = view.lintBuffer.lintResults;
        if (view.lintBuffer.errorString) {
            dialog_alert(view.lintBuffer.errorString);
        } else if (!lintResults) {
            if (view.prefs.getBooleanPref("editUseLinting")) {
                dialog_alert("Syntax checking is currently in progress.");
            } else {
                dialog_alert("Automatic syntax checking disabled: shift-click on the icon to start.");                
            }
        } else if (! lintResults.getNumResults()) {
            dialog_alert("There are no syntax errors.");
        } else {
            // Determine the current position.
            var next = lintResults.getNextResult(view.currentLine,
                                                 view.currentColumn);
            //dump("next lint result: line="+next.lineStart+", column="+
            //     next.columnStart+"\n");

            var pos = view.scimoz.positionAtColumn(next.lineStart-1,
                                                   next.columnStart-1);
            view.scimoz.ensureVisibleEnforcePolicy(next.lineStart-1);
            view.scimoz.gotoPos(pos);
            view.scimoz.selectionStart = pos;
            view.scimoz.selectionEnd = pos;
        }
    } catch(ex) {
        gLintLog.exception(ex);
    }
}


function lint_doRequest() {
    try {
        var view = gViewMgr.currentView;
        view.lintBuffer._clearResults();
        view.lintBuffer._notify();
        view.lintBuffer._issueRequest()
    } catch (e) {
        log.exception(e);
    }
}

function lint_clearResults() {
    try {
        var view = gViewMgr.currentView;
        view.lintBuffer._clearResults();
        StatusBar.ClearCheck(view);
    } catch (e) {
        log.exception(e);
    }
}

function lint_doClick(event) {
    try {
        if (event.shiftKey) lint_doRequest();
    } catch (e) {
        log.exception(e);
    }
}